@@ -15,7 +15,7 @@ extern UV  _XS_prev_prime(UV x);
 extern UV  _XS_prime_count(UV low, UV high);
 extern UV  _XS_nth_prime(UV x);
 
-extern char*  _moebius_range(UV low, UV high);
+extern signed char* _moebius_range(UV low, UV high);
 extern UV*    _totient_range(UV low, UV high);
 extern IV     _XS_mertens(UV n);
 extern double _XS_chebyshev_theta(UV n);
